{-# LANGUAGE FlexibleInstances, FlexibleContexts, UndecidableInstances #-}

module LinPrim (HVec(Point, Vec, Vec4),
                vDot, sDot,
                vAdd, vCross,
                vNeg, vNorm,
                wNorm, vSplat,
                vMag, toTriple,
                toVec, toPoint,
                
                Mat,
                mAdd, mMul,
                vMul, sMul,
                mTranspose, mId,
                mAboutV, mPow,
                mRotPitch, mRotYaw,
                mRotRoll, mRotPYR,
                mvRot, mLook,
                mTranslation, mScale ,
                mProj, mOrth
               )

where
import Numeric
import Data.List (transpose)


--Homogenous coordinate representations
data HVec a = Point a a a
            | Vec a a a
            | Vec4 a a a a
           deriving (Eq)

instance (Num a, Show a) => Show (HVec a) where show = showVec

toCan :: Num a => HVec a -> HVec a
toCan (Point  x y z) = Vec4 x y z 1
toCan (Vec    x y z) = Vec4 x y z 0
toCan (Vec4 x y z w) = Vec4 x y z w

showVec :: (Num a,Show a) => HVec a -> String
showVec = s.toCan
  where
    s (Vec4 x y z w) = "(" ++ show x ++ "," ++ show y ++ "," ++ show z ++ ",[" ++ show w ++  "])"
  
vMag :: Floating a => HVec a -> a
vMag = m.toCan
  where m v = sqrt (v`vDot`v)

toTriple :: Num a => HVec a -> (a,a,a)
toTriple = tri.toCan
  where tri (Vec4 x y z _) = (x,y,z)

toVec :: Num a => HVec a -> HVec a
toVec = tv.toTriple
  where tv (x,y,z) = Vec x y z

toPoint :: Num a => HVec a -> HVec a
toPoint = tp.toTriple
  where tp (x,y,z) = Point x y z

vSplat :: a -> HVec a
vSplat x = Vec x x x

vNorm :: Floating a => HVec a -> HVec a
vNorm v = (recip $ vMag v) `sDot` v

vDot :: Num a => HVec a -> HVec a -> a
vDot l r = d (toTriple l) (toTriple r)
  where d (x, y, z) (x', y', z') = x*x' + y*y' + z*z'
        
v4Dot :: Num a => HVec a -> HVec a -> a
v4Dot l r = d (toCan l) (toCan r)
  where d (Vec4 x y z w) (Vec4 x' y' z' w') = x*x' + y*y' + z*z' + w*w'

sDot :: Num a => a -> HVec a -> HVec a
sDot k = s.toCan
  where s (Vec4 x y z w) = Vec4 (k*x) (k*y) (k*z) w

s4Dot :: Num a => a -> HVec a -> HVec a
s4Dot k = s.toCan
  where s (Vec4 x y z w) = Vec4 (k*x) (k*y) (k*z) (k*w)

vAdd :: (Num a, Ord a) => HVec a -> HVec a -> HVec a
vAdd l r = a (toCan l) (toCan r)
  where a (Vec4 x y z w) (Vec4 x' y' z' w') = Vec4 (x+x') (y+y') (z+z') (max w w')

v4Add :: Num a => HVec a -> HVec a -> HVec a
v4Add l r = a (toCan l) (toCan r)
  where a (Vec4 x y z w) (Vec4 x' y' z' w') = Vec4 (x+x') (y+y') (z+z') (w + w') 

vNeg :: Num a => HVec a -> HVec a
vNeg = n.toCan
  where n (Vec4 x y z w) = Vec4 (-x) (-y) (-z) w

vCross :: Num a => HVec a -> HVec a -> HVec a
vCross l r = c (toTriple l) (toTriple r)
  where c (x, y, z) (x', y', z') = Vec (y*z' - z*y') (z*x' - x*z') (x*y' - y*x')

wNorm :: Floating a => HVec a -> HVec a
wNorm = w.toCan
  where w (Vec4 x y z w) = Point (x/w) (y/w) (z/w)


--Matricies
data Mat a = Mat (HVec a) (HVec a) (HVec a) (HVec a)

instance {-# Overlappable #-} Show a => Show (Mat a) where
  show = showMat show
  
instance {-# Overlapping #-} Show (Mat Float) where
  show = showMat roundShow

instance {-# Overlapping #-} Show (Mat Double) where
  show = showMat roundShow

roundShow :: (Ord a, RealFloat a) => a -> String
roundShow x = (if x >= 0 then " " else "") ++ showFFloat (Just 2) x ""

--Layout matrix as a table
showMat :: (a->String) -> Mat a -> String
showMat s (
  Mat (Vec4 a b c d)
      (Vec4 e f g h)
      (Vec4 i j k l)
      (Vec4 m n o p)
  ) = layout [
  ["/",  s a, " ", s b, " ", s c, " ", s d, " \\"],
  ["|",  s e, " ", s f, " ", s g, " ", s h, " |"],
  ["|",  s i, " ", s j, " ", s k, " ", s l, " |"],
  ["\\", s m, " ", s n, " ", s o, " ", s p, " /"]
  ]

--layout strings to a centred table output string
layout :: [[String]] -> String
layout = unlines.tab.transpose
  where
    tab = (map concat).transpose.pad 
    pad xss = map (\xs-> map (padTo (widest xs)) xs) xss
    widest = maximum.(map length)
    padTo n str = (replicate (((n-length str)+1)`div`2) ' ') ++ str ++ (replicate ((n-length str)`div`2) ' ')

mAdd :: (Ord a, Num a) => Mat a -> Mat a -> Mat a
mAdd (Mat r1 r2 r3 r4) (Mat r1' r2' r3' r4')
  = Mat (r1 `v4Add` r1')
        (r2 `v4Add` r2')
        (r3 `v4Add` r3')
        (r4 `v4Add` r4') 

mTranspose :: Mat a -> Mat a
mTranspose (
  Mat (Vec4 a b c d)
      (Vec4 e f g h)
      (Vec4 i j k l)
      (Vec4 m n o p)
  ) =
  Mat (Vec4 a e i m)
      (Vec4 b f j n)
      (Vec4 c g k o)
      (Vec4 d h l p)

mMul :: Num a => Mat a -> Mat a -> Mat a
mMul (Mat r1 r2 r3 r4) m
  = Mat (Vec4 (r1`v4Dot`c1') (r1`v4Dot`c2') (r1`v4Dot`c3') (r1`v4Dot`c4'))
        (Vec4 (r2`v4Dot`c1') (r2`v4Dot`c2') (r2`v4Dot`c3') (r2`v4Dot`c4'))
        (Vec4 (r3`v4Dot`c1') (r3`v4Dot`c2') (r3`v4Dot`c3') (r3`v4Dot`c4'))
        (Vec4 (r4`v4Dot`c1') (r4`v4Dot`c2') (r4`v4Dot`c3') (r4`v4Dot`c4'))
  where (Mat c1' c2' c3' c4') = mTranspose m

vMul :: Num a => Mat a -> HVec a -> HVec a
vMul (Mat r1 r2 r3 r4) v = Vec4 (r1`v4Dot`v) (r2`v4Dot`v) (r3`v4Dot`v) (r4`v4Dot`v)

sMul :: Num a => a -> Mat a -> Mat a
sMul k (Mat r1 r2 r3 r4) = Mat (k`sDot`r1) (k`sDot`r2) (k`sDot`r3) (k`sDot`r4)

mPow :: (Num a, Integral b) => Mat a -> b -> Mat a
mPow x n = pow x mId n
  where
    pow m x 0 = x
    pow m x n = if odd n
                then pow (m`mMul`m) (x`mMul`m) (n`div`2)
                else pow (m`mMul`m) x (n`div`2)

--Useful matrices

mId :: Num a => Mat a
mId = Mat (Vec4 1 0 0 0)
          (Vec4 0 1 0 0)
          (Vec4 0 0 1 0)
          (Vec4 0 0 0 1)

mRotPitch :: Floating a => a -> Mat a
mRotPitch theta = Mat (Vec4 1      0           0               0)
                      (Vec4 0 (cos theta) (negate $ sin theta) 0)
                      (Vec4 0 (sin theta) (cos theta)          0)
                      (Vec4 0      0           0               1)

mRotYaw :: Floating a => a -> Mat a
mRotYaw theta = Mat (Vec4 (cos theta)          0 (sin theta) 0)
                    (Vec4      0               1      0      0)
                    (Vec4 (negate $ sin theta) 0 (cos theta) 0)
                    (Vec4      0               0      0      1)

mRotRoll :: Floating a => a -> Mat a
mRotRoll theta = Mat (Vec4 (cos theta) (negate $ sin theta) 0 0)
                     (Vec4 (sin theta) (cos theta)          0 0)
                     (Vec4       0          0               1 0)
                     (Vec4       0          0               0 1)

mRotPYR :: Floating a => a -> a -> a -> Mat a
mRotPYR p y r = (mRotYaw y)`mMul`(mRotPitch p)`mMul`(mRotRoll r)

mvRot :: (Ord a,Floating a) => HVec a -> Mat a
mvRot b
  | s == 0 = if (a == b) then mId else mScale (Vec 1 1 (-1))
  |otherwise = mId `mAdd` skew `mAdd` (((1-c)/(s*s)) `sMul` (skew `mMul` skew))
  where
  a = (Vec 0 0 1)
  v = a `vCross` b
  s = vMag v
  c = a `vDot` b
  (x,y,z) = toTriple v
  skew = Mat (Vec4   0 (-z)  y  0)
             (Vec4   z   0 (-x) 0)
             (Vec4 (-y)  x   0  0)
             (Vec4   0   0   0  0)

mAboutV :: (Ord a,Floating a) => HVec a -> Mat a -> Mat a
mAboutV a m | xyLen == 0 = m
            |otherwise=(mvRot a) `mMul` m `mMul` t
  where
    (x,y,z) = toTriple $ vNorm a
    xyLen = sqrt (x*x + y*y)
    t = t2 `mMul` t1
    t1 = Mat (Vec4 ( x/xyLen) (y/xyLen) 0 0)
             (Vec4 (-y/xyLen) (x/xyLen) 0 0)
             (Vec4       0         0    1 0)
             (Vec4       0         0    0 1)
         
    t2 = Mat (Vec4   z   0 (-xyLen) 0)
             (Vec4   0   1     0    0)
             (Vec4 xyLen 0     z    0)
             (Vec4   0   0     0    1)
 
mLook :: (Ord a, Floating a) => HVec a -> Mat a
mLook v
  = Mat xAxis yAxis zAxis (Vec4 0 0 0 1)
    where
      zAxis = toCan $ vNorm $ vNeg v
      xAxis = toCan $ vCross (Vec 0 1 0) zAxis
      yAxis = toCan $ vCross zAxis xAxis

mTranslation :: Num a => HVec a -> Mat a 
mTranslation = t.toCan
  where t (Vec4 x y z _) = Mat (Vec4 1 0 0 x)
                               (Vec4 0 1 0 y)
                               (Vec4 0 0 1 z)
                               (Vec4 0 0 0 1)

mScale :: Num a => HVec a -> Mat a
mScale (Vec x y z)
  = Mat (Vec4 x 0 0 0)
        (Vec4 0 y 0 0)
        (Vec4 0 0 z 0)
        (Vec4 0 0 0 1)

mProj :: Floating a => a -> a -> a -> a -> Mat a
mProj hFov vFov near far
  = Mat (Vec4 wBn 0 0 0)
        (Vec4 0 hBn 0 0)
        (Vec4 0   0 a b)
        (Vec4 0   0 1 0)
  where
    wBn = 1/(tan (hFov/2))
    hBn = 1/(tan (vFov/2))
    a = (far + near) / (far-near)
    b = (2*near*far) / (far-near)

mOrth :: Fractional a => a -> a -> a -> a -> Mat a
mOrth width height near far
  = Mat (Vec4 (2/width) 0      0 0)
        (Vec4 0     (2/height) 0 0)
        (Vec4 0         0      a b)
        (Vec4 0         0      0 1)
    where
      a = 2 / (near - far)
      b = (near + far) / (near - far)
 
