import Control.Monad.State
import Control.Monad
import Data.Monoid ((<>))
import Data.List (isSuffixOf)
import System.Environment (getArgs)
import LinPrim
import Render
import Colour
import Vert
import Camera
import Lights
import Utils

main :: IO ()
main = do
  args <- getArgs
  n <- return (if null args
               then 3
               else read $ head args)
  tree <- theTree n
  renderObjs ("../tree - " ++ show n) scene tree
  where
    scene = Scene (Lights white [(DirectL (vNorm (Vec 0.3 (-5.0) 0.5)) sunshine (0.6*white)),
                                 (DirectL (vNorm (Vec (-0.1) 0.0 (-1.0))) (0.3*white) (0.8*white))
                                      ])
                  (Cam (Point (3.0) (12.0) (-20.0)) (Vec (0.0) (-0.1) (1.0)) (radFromDeg 2.5)
                   (mkPersp 1920 1080 (radFromDeg 80) 0.1 100))
    theTree n = do
      wmat <- woodMat
      wmesh <- wood
      lmat <- leafMat
      lmesh <- leaves
      return $ map (applyOMat scale) $ tree (Object wmat wmesh) (Object lmat lmesh) n 
      where
        scale = mScale $ vSplat (1.2/(2^n))

circle :: Int -> [Vert]
circle n = map (\p -> (p,toVec p,toVec p)) points
  where
    point theta = (Point (cos theta) (sin theta) 0.0)
    points = map (point.((2*pi/(fromIntegral n))*).fromIntegral) [0..n-1]

filledCircle :: Int -> Mesh
filledCircle n = mkMesh vs is
  where
    vs = map (\(p,n,t)->(p,Vec 0.0 0.0 (-1.0),t)) $ circle n
    is = concat [[0, a, a + 1]| a <- [1..(n-2)]]
    
tube :: Int -> Mesh
tube n = mkMesh vs is
  where
    zShift a (p,n,c) = (p `vAdd` (Vec 0.0 0.0 a) , n, c)
    vs = (circle n) ++ (map (zShift 8) (circle n))
    is = (concatMap (\m -> [m, (n + m), ((m+1) `mod` n),
                           ((m+1) `mod` n), (n + m), (n + ((m+1) `mod` n))])
         [0..(n-1)])

cylinder :: Int -> Mesh
cylinder n = (tube n) <>
             (filledCircle n) <>
             (applyMMat flip (filledCircle n))
  where
    flip = (mTranslation (Vec 0 0 8))`mMul`(mRotPitch pi)
    
icoSphere :: IO (Mesh)
icoSphere = meshFromObjFile "meshs/icosphere.obj"


type TurtleInf = (HVec Float, HVec Float) --position, facing vector



tree :: Object -> Object -> Int -> [Object]
tree branch leaf n
  = ret $ execState (x n) (( Point 0.0 (0.0) 0.0, Vec 0.0 1.0 0.0),
                           (Object (objMaterial branch) mempty, Object (objMaterial leaf) mempty ))
  where
    ret (_,(a,b)) = [a,b]
    x :: Int -> State (TurtleInf, (Object, Object)) ()
    x 0 = modify (\((pos,dir), (b,l))->(((pos`vAdd`(7.9`sDot`dir), dir), (b <++ (newBranch 0 pos dir), l))))
          >> addLeaf
    x n = do
      f (n-1)
      (t,_) <- get
      rot ((r `mPow` 3) `mMul` y `mMul` p')
      x (n-1)
      setTurtle t
      rot ((r `mPow` 7) `mMul` y `mMul` p')
      x (n-1)
      setTurtle t
      rot ((r `mPow` 11) `mMul` y `mMul` p')
      x (n-1)
      setTurtle t
      f (n-1)
      x (n-1)
      addLeaf
      
    f :: Int -> State (TurtleInf, (Object, Object)) ()
    f 0 = x 0
    f n = modify (\((pos,dir), (b,l))->(((pos`vAdd`(((2^n)*7.9)`sDot`dir), dir),
                                         (b <++ (newBranch (fromIntegral n) pos dir), l))))
      
    addLeaf = modify (\((pos,dir), (b, l))->(((pos, dir),(b, l <++ (newLeaf pos dir)))))

    setTurtle :: TurtleInf -> State (TurtleInf, (Object, Object)) ()
    setTurtle t = modify (\(t',m')->(t,m'))
    
    rot :: Mat Float -> State (TurtleInf, (Object, Object)) ()
    rot mat = modify (\((p,d),t)->((p,vNorm((mAboutV d mat)`vMul`d)),t))

    a  = radFromDeg 30
    y  = mRotYaw   ( a)
    y' = mRotYaw   (-a)
    p  = mRotPitch ( a)
    p' = mRotPitch (-a)
    r  = mRotRoll  ( a)
    r' = mRotRoll  (-a)
    
    newLeaf p d = applyOMat ((mTranslation p)`mMul`(mvRot d)`mMul`mScale (vSplat $ log $ fromIntegral (n+2))) (leaf)
    newBranch s p d = applyOMat ((mTranslation p)`mMul`(mvRot d)`mMul`
                            mScale (Vec ((s+2)/2) ((s+2)/2) (2^(round s)))) branch

leafMesh :: IO (Mesh)
leafMesh = meshFromObjFile "meshs/leaf.obj"

leaves :: IO (Mesh)
leaves = do
  leaf <- leafMesh
  return $ mconcat $ ls leaf
  where ls l = [applyMMat (mScale (vSplat 0.7)) l,
                applyMMat (mScale (vSplat 0.4)`mMul`mvRot (Vec   0.8  0.0 0.8)) l,
                applyMMat (mScale (vSplat 0.4)`mMul`mvRot (Vec (-0.8) 0.0 0.8)) l]

leafMat :: IO (Material)
leafMat = mkMaterial "textures/leaf-lr.png" 8 (Colour 0.7 1.0 0.5) (Colour 0.4 0.5 0.5) (Colour 0.4 0.4 0.4)

testLeaf :: IO ()
testLeaf = do
  leaf <- leafMesh
  lmat <- leafMat
  testRenderObjs "test/Leaf" [(Object lmat (applyMMat (mRotPitch (-pi/2)) leaf))]

wood :: IO (Mesh)
wood = meshFromObjFile "meshs/branch.obj"

woodMat :: IO (Material)
woodMat = mkMaterial "textures/bark.png" 2 (0.6*white) brown (0.4*white)


testTree :: Int -> IO ()
testTree n = do
  tree <- theTree
  meshToObjFile "test/tree" (mconcat $ map (\(Object _ m)->m) tree)
  testRenderObjs "test/tree" tree
  where
    theTree = do
      lmat <- flatMaterial (0.4*green) 6 (Colour 0.5 0.6 0.3) (Colour 0.8 0.7 0.75) (Colour 0.6 0.6 0.6)
      lmesh <- leaves
      tmat <- testMaterial
      return $ map (applyOMat (move`mMul`scale)) $
        tree (Object tmat (cylinder 8)) (Object lmat lmesh) n 
    scale = mScale $ vSplat (1.2/(2^n))
    move = mTranslation (Vec 0.0 (-5.0) 0.0)
              
testSquares :: IO ()
testSquares = --sequence_ $ zipWith meshToObjFile ["sq1", "sq2"] squares
  testRenderMeshs "test/WSquare" squares
  where squares = [applyMMat (mTranslation (Vec 0.0 0.5 (5.0))`mMul`mScale (Vec 4.0 4.0 1.0)) square,
                   applyMMat (mRotPYR 0.5 0.5 0.0 `mMul` mScale (Vec 2.0 2.0 1.0)) square]
        square = mkMesh vs is
        vs = [(Point 0.5    0.5    0.0   ,Vec 0.0 0.0 (1.0),Vec 1.0 1.0 0.0),
              (Point (-0.5) 0.5    0.0   ,Vec 0.0 0.0 (1.0),Vec 0.0 1.0 0.0),
              (Point (-0.5) (-0.5) 0.0   ,Vec 0.0 0.0 (1.0),Vec 0.0 0.0 0.0),
              (Point 0.5    (-0.5) 0.0   ,Vec 0.0 0.0 (1.0),Vec 1.0 0.0 0.0)
             ]
        is = [0, 1, 2,   0, 2, 3]
        
testCylinder :: IO ()
testCylinder = testRenderMeshs "test/Cylinder" [m]
  where m = applyMMat (translate `mMul` rotate) (cylinder 8)
        translate = mTranslation (Vec 0.0 (4.0) 0.0)
        rotate = mRotPitch (0.5-(3*pi/2))

testSphere :: IO ()
testSphere = do
  sph <- icoSphere
  testRenderMeshs "test/sphere" [applyMMat ((mScale $ vSplat 5.0) `mMul` mRotYaw 0.2) sph]

testScene = Scene (Lights white [(DirectL (vNorm (Vec 0.3 (-5.0) 0.5)) sunshine (white * 0.8)),
                                 (DirectL (vNorm (Vec (-0.1) 0.0 (-1.0))) (white * 0.6) white)
                                      ])
                  (Cam (Point (0.0) (4.0) (-20.0)) (Vec (0.0) (0.0) (1.0)) 0 (mkPersp 1920 1080 (radFromDeg 80) 0.1 100))

testMaterial = flatMaterial brown 2 (Colour 0.3 0.5 0.3) (Colour 0.8 0.7 0.75) (Colour 0.6 0.6 0.6)

renderObjs :: String -> Scene -> [Object] -> IO ()
renderObjs fn sc objs = do
  buffs <- createBuffs 1920 1080 cornflowerBlue
  mapM_ (render buffs) $ map (applyOCam (scCam sc)) objs
  join $ save fn buffs
  where
    render buffs obj = drawTris buffs sc obj

testRenderObjs :: String -> [Object] -> IO ()
testRenderObjs fn = renderObjs fn testScene

testRenderMeshs :: String -> [Mesh] -> IO()
testRenderMeshs fn meshs = do
  material <- testMaterial
  testRenderObjs fn (map (Object material) meshs)
