module Render
       (save, createBuffs,
        drawLine, drawLines,
        drawTri, drawTris

       )
       where
import Codec.Picture
import Codec.Picture.Types
import Control.Monad
import Control.Monad.Primitive
import Data.List (isSuffixOf, sortBy)
import qualified Data.Vector as V
import LinPrim
import Camera
import Utils
import Lights
import Colour
import Vert

type RGB  = PixelRGB8
type Scal = PixelF

type IBuffer m = MutableImage (PrimState m) RGB
type DBuffer m = MutableImage (PrimState m) Scal

type Buffers m = (IBuffer m, DBuffer m)
type PixelData = (RGB, Scal)

group :: Monad m => m a -> m b -> m (a,b)
group = liftM2 mkPair
  where mkPair a b = (a,b) 

save :: PrimMonad m => FilePath -> Buffers m -> m (IO ())
save path (iB, dB) = do
  img <- freezeImage iB
  return (savePngImage (fileExtension ".png" path) (ImageRGB8 img))
 
createBuffs :: PrimMonad m => Int -> Int -> Colour -> m (Buffers m)
createBuffs w h bg = do group iBuf dBuf
    where
      iBuf = createMutableImage w h (toPixel bg)
      dBuf = createMutableImage w h 1.0
  

readBuffs :: PrimMonad m =>  Buffers m -> Int -> Int -> m PixelData
readBuffs (i, d) x y = group (rp i) (rp d)
  where rp b = readPixel b x y

writeBuffs :: PrimMonad m => Buffers m -> Int -> Int -> PixelData -> m ()
writeBuffs (iB, dB) x y (p', d')= do {(wp iB p') ; (wp dB d')}
  where wp b p = writePixel b x y p

updateBuffs :: PrimMonad m => Buffers m -> Int -> Int -> PixelData -> m ()
updateBuffs bs x y (p', d')
  | d' > (-1) = do
      (_, d) <- readBuffs bs x y
      if (d > d')
        then writeBuffs bs x y (p', d')
        else return ()
  | otherwise = return ()


line :: Int -> Int -> Scene -> Material -> Vert -> Vert -> [RastAct]
line w h sc mat p q = [shadeL sc mat v |
                       v <- map interp $ enumBtw x x',
                       yOf v == clpRnd( getY v ) ]
  where
    yOf = round.getY
    (x, x') = (clpRnd $ getX p, clpRnd $ getX q)
    clpRnd a = clip 0 (w-1) (round a)
    interp b = interpolX (fromIntegral b) p q


drawLine :: PrimMonad m => Buffers m -> Scene -> Material -> Vert -> Vert -> m ()
drawLine bs@(MutableImage w h _, _) sc mat v0 v1 =
  mapM_ (set.(octZeroTo w h relOct))
  ((l relOct) (apply2 (octZeroFrom relOct) (v0, v1)))
  where
    --handles when transforming to 0th octant requires
    --different raster space
    l n (a,b) = case n of
                     1 -> line h w sc mat (toRasterSpace h w a) (toRasterSpace h w b)
                     2 -> line h w sc mat (toRasterSpace h w a) (toRasterSpace h w b)
                     5 -> line h w sc mat (toRasterSpace h w a) (toRasterSpace h w b)
                     6 -> line h w sc mat (toRasterSpace h w a) (toRasterSpace h w b)
                     _ -> line w h sc mat (toRasterSpace w h a) (toRasterSpace w h b)
    apply2 f (a,b) = (f a, f b)
    relOct = octOf v0 v1
    set (x, y, z, c) = updateBuffs bs x y (c, z)

drawLines :: PrimMonad m => Buffers m -> Scene -> Object -> m ()
drawLines bs sc (Object mat mesh)
  = mapLineM_ (drawLine bs sc mat) mesh

--draw tri where v1.y == v2.y
drawAlignedTri :: PrimMonad m => Buffers m -> Scene -> Material -> Vert -> Vert -> Vert -> m()
drawAlignedTri bs@(MutableImage w h _,_) sc mat v0 v1 v2
  = mapM_ draw (enumBtw y0 y1)
  where
    draw y = mapM_ set (line w h sc mat
                        (interpolY (fromIntegral y) r2 r0)
                        (interpolY (fromIntegral y) r1 r0)
                       )
    clpRnd a = clip 0 (h-1) (round a)
    r0 = toRasterSpace w h v0
    r1 = toRasterSpace w h v1
    r2 = toRasterSpace w h v2
    set (x, y, z, c) = updateBuffs bs x y (c, z)
    y0 = clpRnd $ getY r0
    (x1,y1) = (getX r1, clpRnd $ getY r1)
    x2 = getX r2

 
drawTri :: PrimMonad m => Buffers m -> Scene -> Material -> Vert -> Vert -> Vert -> m()
drawTri bs sc mat v0 v1 v2
  | cull = return ()
  | otherwise = do
      drawAlignedTri bs sc mat oV0 newVert oV1
      drawAlignedTri bs sc mat oV2 newVert oV1
  where
    (oV0:oV1:oV2:[]) = sortBy compareY [v0,v1,v2]
    midY = getY oV1
    compareY a b = compare (getY a) (getY b)
    newVert = interpolY midY oV0 oV2
    (x0,y0) = (getX v0, getY v0)
    (x1,y1) = (getX v1, getY v1)
    (x2,y2) = (getX v2, getY v2)
    cull = (x0*(y1-y2) + x1*(y2-y0) + x2*(y0-y1)) >= 0

drawTris :: PrimMonad m => Buffers m -> Scene -> Object -> m ()
drawTris bs sc (Object mat mesh)
  = mapTriM_ (drawTri bs sc mat) mesh






--------------
--Test draws--
--------------

testScene = Scene (Lights white [(DirectL (vNorm (Vec 0.3 (-1.0) 0.5)) (Colour 0.6 0.2 0.2) white)])
                  (Cam (Point 0.0 (1.0) (-4.0)) (Vec 0.0 (-0.5) 1.0) 0 (mkPersp 640 480 1.3 0.1 100))

testMaterial = flatMaterial red 4 (Colour 0.8 0.8 0.8) (Colour 0.8 0.7 0.75) (Colour 0.6 0.6 0.6)

testLine :: IO ()
testLine = do
  buffs <- createBuffs 640 480 cornflowerBlue
  material <- testMaterial
  drawLine buffs testScene material
    (Point 0.0 (-0.5) 0.5,Vec 0.0 0.0 (-1.0),Vec 0.0 0.0 0.0)
    (Point (-0.5) 0.5 0.5,Vec 0.0 0.0 (-1.0),Vec 0.0 0.0 0.0)
  join $ save "test/Line" buffs

--draw triangle with top in line with raster grid
testAlignedTri :: IO ()
testAlignedTri = do
  buffs <- createBuffs 640 480 cornflowerBlue
  material <- testMaterial
  drawAlignedTri buffs testScene material
    (Point 0.0    0.0    0.0,Vec 0.0 0.0 (-1.0),Vec 0.0 0.0 0.0)
    (Point 0.5    (-0.5) 0.0,Vec 0.0 0.0 (-1.0),Vec 0.0 0.0 0.0)
    (Point (-0.5) (-0.5) 0.0,Vec 0.0 0.0 (-1.0),Vec 0.0 0.0 0.0)
  join $ save "test/AlignedTri" buffs

--draw triangle not aligned to a raster grid
testTri :: IO()
testTri = do
  buffs <- createBuffs 640 480 cornflowerBlue
  material <- testMaterial
  drawTri buffs testScene material
    (0.5 `sDot` (Point (cos (theta-pi/3))   (sin (theta-pi/3))   0.5),Vec 0.0 0.0 (-1.0),Vec 0.0 0.0 0.0)
    (0.5 `sDot` (Point (cos (theta-2*pi/3)) (sin (theta-2*pi/3)) 0.5),Vec 0.0 0.0 (-1.0),Vec 0.0 0.0 0.0)
    (0.5 `sDot` (Point (cos (theta+pi/2))   (sin (theta+pi/2))   0.5),Vec 0.0 0.0 (-1.0),Vec 0.0 0.0 0.0)
  join $ save "test/Tri" buffs
  where theta = 0.5

--draw using index
testSquare :: IO ()
testSquare = do
  buffs <- createBuffs 640 480 cornflowerBlue
  material <- testMaterial
  drawTris buffs testScene (Object material
    (mkMesh [((Point 0.5    0.5    0.5),Vec 0.0 0.0 (-1.0),Vec 0.0 0.0 0.0),
             ((Point (-0.5) 0.5    0.5),Vec 0.0 0.0 (-1.0),Vec 0.0 0.0 0.0),
             ((Point (-0.5) (-0.5) 0.5),Vec 0.0 0.0 (-1.0),Vec 0.0 0.0 0.0),
             ((Point 0.5    (-0.5) 0.5),Vec 0.0 0.0 (-1.0),Vec 0.0 0.0 0.0)
            ]
     [2, 1, 0, 3, 2, 0])
    )
  join $ save "test/Square" buffs

  
