module Vert
       ( Vert(..), RastAct,
         getX, getY,
         interpolX, interpolY,
         interpol,
         toRasterSpace,
         octZeroFrom, octZeroTo,
         octOf, shadeL,

         Mesh(..), Object(..),
         mkMesh, (<++),
         applyMCam, applyMMat,
         applyOCam, applyOMat,
         mapLineM_, mapTriM_,
         mapPos, mapNorm, mapTex,
         meshToObjFile,
         meshFromObjFile
       )
       where
import Codec.Picture (pixelAt)
import Codec.Picture.Types (Image(Image), PixelRGB8)
import qualified Data.Vector as V
import Data.List (dropWhile)
import Camera
import LinPrim
import Utils
import Colour
import Lights


--Vertex  (Pos, Norm, Tex) --tex is assumed (Vec u v _ _) 
type Vert = (HVec Float, HVec Float, HVec Float)

interpol :: Vert -> Vert -> Float -> Vert
interpol (v0,n0,t0) (v1,n1,t1) s = (toPoint $ vLerp v0 v1 s',
                                      toVec $ vLerp n0 n1 s',
                                      toVec $ vLerp t0 t1 s')
  where
    s' = clip 0 1 s

--find the intersection between the line from v0-v1 and the y value given
interpolY :: Float -> Vert -> Vert -> Vert
interpolY y v0 v1 = interpol v0 v1 ratio
  where
    (y0, y1) = (getY v0, getY v1)
    ratio = ((y-y0)/(y1-y0))
    
--find the intersection between the line from v0-v1 and the x value given
interpolX :: Float -> Vert -> Vert -> Vert
interpolX x v0 v1 = interpol v0 v1 ratio
  where
    (x0, x1) = (getX v0, getX v1)
    ratio = ((x-x0)/(x1-x0))

--convert from normalised coords in range ([-1,1],[-1,1],[-1,1]) to ([0,w),[0,h),[-1,1])
toRasterSpace :: (Floating a, RealFrac a) => Int -> Int -> Vert -> Vert
toRasterSpace w h (p,n,t)
  = (scale`vMul`(p`vAdd`(Vec 1 1 0)),n,t)
  where
    tW = (fromIntegral (w - 1)) / 2
    tH = (fromIntegral (h - 1)) / 2
    scale = mScale (Vec tW tH 1)


octZeroTo :: Floating a => Int -> Int -> Int -> RastAct -> RastAct
octZeroTo w h n (x,y,z,c)
  = case n of
     0 -> (x   ,y   ,z,c)
     1 -> (y   ,x   ,z,c)
     2 -> (w'-y,x   ,z,c)
     3 -> (w'-x,y   ,z,c)
     4 -> (w'-1,h'-y,z,c)
     5 -> (w'-y,h'-x,z,c)
     6 -> (y   ,h'-x,z,c)
     7 -> (x   ,h'-y,z,c)
  where
    w' = w - 1
    h' = h - 1
                     
                     
octZeroFrom :: Floating a => Int -> Vert -> Vert
octZeroFrom oct (p,n,c) = case oct of
                           0 -> (Point x    y    z,n,c)
                           1 -> (Point y    x    z,n,c)
                           2 -> (Point y    (-x) z,n,c)
                           3 -> (Point (-x) y    z,n,c)
                           4 -> (Point (-x) (-y) z,n,c)
                           5 -> (Point (-y) (-x) z,n,c)
                           6 -> (Point (-y)  x   z,n,c)
                           7 -> (Point x    (-y) z,n,c)
  where
    (x,y,z) = toTriple p

octOf :: (Ord a, Floating a) => Vert -> Vert -> Int
octOf (p0,_,_) (p1,_,_) = o $ toTriple (p1 `vAdd` (vNeg p0))
  where
    o :: (Ord a, Floating a) => (a, a, a) -> Int
    o (dX, dY, _) =
      if dX > 0 then
        if dY > 0 then 
          if dY > dX then 1
          else 0
        else
          if (-dY) > dX then 6
          else 7
      else
        if dY > 0 then
          if dY > (-dX) then 2
          else 3
      else
          if dY < dX then 5
          else 4

getX :: Vert -> Float
getX (p,_,_) = x
  where (x,_,_) = toTriple p

getY :: Vert -> Float
getY (p,_,_) = y
  where (_,y,_) = toTriple p

getTexel :: Material -> HVec Float -> Colour
getTexel (Material tex@(Image w h _) _ _ _ _) t = fromPixel $ pixelAt tex x' y'
  where inRange a = a >= 0 && a <= 1
        (x,y,_) = toTriple t
        x' = round (x*(fromIntegral (w-1))) `mod` w
        y' = round ((1-y)*(fromIntegral (h-1))) `mod` h

shadeL :: Scene -> Material -> Vert -> RastAct
shadeL (Scene (Lights aL dLs) (Cam pos look _ _))
      m@(Material _ shine spec diff amb) (p, n, t)
 = (round x, round y,z,toPixel newC)
   where
     c = getTexel m t
     newC = c * (aL * amb + (sum $ map dirLight dLs))
     look' = vNorm $ vNeg look
     dirLight (DirectL dir dLum sLum) = phong shine (dLum*diff) (sLum*spec) (vNeg dir) n look'
     (x,y,z) = toTriple p

-- shine, diffuse luminance, specular luminance, light, normal, viewer
phong :: Int -> Colour -> Colour -> HVec Float -> HVec Float -> HVec Float -> Colour
phong shine dLum sLum l n v = diffuse + specular
  where
    diffuse = (cSplat (l`vDot`n))*dLum
    specular = (cSplat ((r`vDot`v)^^shine))*sLum
    r = ((2*(l`vDot`n))`sDot`n) `vAdd` (vNeg l)



--Compose vertices to create meshes and objects

data Mesh = Mesh {meshPoints ::(V.Vector (HVec Float)),
                  meshNorms :: (V.Vector (HVec Float)),
                  meshTexs ::(V.Vector (HVec Float)),
                  meshIndices :: [(Int,Int,Int)]}
          deriving (Eq, Show)
data Object = Object {objMaterial :: Material, objMesh :: Mesh}
            deriving (Show)

instance Monoid Mesh where
  mempty = Mesh (V.empty) (V.empty) (V.empty) []
  mappend = meshAppend

(<++) :: Object -> Object -> Object
(Object mat mesh) <++ (Object _ mesh')
  = Object mat (mesh `mappend` mesh')

mkMesh :: [Vert] -> [Int] -> Mesh
mkMesh vs is
  | all inRange is = Mesh (V.fromList ps) (V.fromList ns)
                     (V.fromList ts) (map (\i->(i,i,i)) is)
  | otherwise      = error "Mesh index out of range"
  where
    (ps,ns,ts) = unzip3 vs
    inRange x = x >= 0 && x < l
    l = length vs

meshAppend :: Mesh -> Mesh -> Mesh
meshAppend (Mesh ps ns ts is) (Mesh ps' ns' ts' is')
  = Mesh (ps V.++ ps') (ns V.++ ns') (ts V.++ ts') (is ++ (map adj is'))
  where adj (pi,ni,ti) = (pi + psLen, ni + nsLen, ti + tsLen)
        psLen = V.length ps
        nsLen = V.length ns
        tsLen = V.length ts

mapPos :: (HVec Float -> HVec Float) -> Mesh -> Mesh
mapPos f (Mesh ps ns ts is) = Mesh (V.map f ps) ns ts is

mapNorm :: (HVec Float -> HVec Float) -> Mesh -> Mesh
mapNorm f (Mesh ps ns ts is) = Mesh ps (V.map f ns) ts is

mapTex :: (HVec Float -> HVec Float) -> Mesh -> Mesh
mapTex f (Mesh ps ns ts is) = Mesh ps ns (V.map f ts) is

mkVert :: Mesh -> (Int,Int,Int) -> Vert
mkVert (Mesh ps ns ts _) (pi,ni,ti) = ((ps V.! pi), (ns V.! ni), (ts V.! ti))

mapLineM_ :: Monad m => (Vert -> Vert -> m a) -> Mesh -> m ()
mapLineM_ f m@(Mesh ps ns ts is) = ml is
  where ml (a:b:xs) = (f (mkVert m a) (mkVert m b)) >> ml xs
        ml xs = return ()
        
mapTriM_ :: Monad m => (Vert -> Vert -> Vert -> m a) -> Mesh -> m ()
mapTriM_ f m@(Mesh ps ns ts is) = mt is
  where mt (a:b:c:xs) = (f (mkVert m a) (mkVert m b) (mkVert m c)) >>  mt xs
        mt xs = return ()

applyMMat :: Mat Float -> Mesh -> Mesh
applyMMat mat (Mesh ps ns ts is) = Mesh (V.map (vMul mat) ps) (V.map (vMul mat) ns) ts is

applyMCam :: Camera Float -> Mesh -> Mesh
applyMCam cam = (mapPos wNorm) . (applyMMat (camMat cam))

applyOMat :: Mat Float -> Object -> Object
applyOMat m (Object mat mesh) = Object mat $ applyMMat m mesh

applyOCam :: Camera Float -> Object -> Object
applyOCam cam (Object mat mesh) = Object mat $ applyMCam cam mesh 

meshToObjFile :: FilePath -> Mesh -> IO ()
meshToObjFile fn (Mesh ps ns ts is)
  = writeFile (fileExtension ".obj" fn) (vs ++ vts ++ vns ++ (fs is))
  where
    cat3 sep (a,b,c) = a ++ sep ++ b ++  sep ++ c
    show3 sep (a,b,c) = cat3 sep (show a, show b, show c)
    show2of3 sep (a,b,_) = show a ++ sep ++ show b
    
    vs = unlines $ V.toList $ V.map
         (\p-> "v " ++ (show3 " " $ toTriple p)) ps
    vts = unlines $ V.toList $ V.map
          (\t -> "vt " ++ (show2of3 " " $ toTriple t)) ts
    vns = unlines $ V.toList $ V.map
          (\n -> "vn " ++ (show3 " "$ toTriple n)) ns
    idx (a,b,c) = show3 "/" (a+1,b+1,c+1)
    toFace a b c = "f " ++ cat3 " " (idx a, idx b, idx c)
    fs (a:b:c:xs) = toFace a b c ++ "\n" ++ fs xs
    fs _ = ""

meshFromObjFile :: FilePath -> IO (Mesh)
meshFromObjFile fn = do
  file <- readFile fn
  return $ resolve $ parse $ lines file
  where
    --([HVec Float], [HVec Float], [HVec Float], [(Maybe Int,Maybe Int,Maybe Int)])
    --positions, textures, normals, indices (posIdx, texIdx, normIdx)
    parse = foldr update ([], [], [], [])
    update (c:cs) = case c of
                     'f' -> (\x->foldr parseFace x $ (map (split '/') . split ' ' . dropWhile (==' ')) cs)
                     'v' -> parseVert cs
                     _   -> id
    mbRead "" = Nothing
    mbRead str = Just $ read str
    jPred x = Just (x-1)
    parseFace [s] = let x = (mbRead s >>= jPred) in addIndex (x,x,x)
    parseFace [p,t] = addIndex (mbRead p >>= jPred, mbRead t >>= jPred, Nothing)
    parseFace [p,t,n] = addIndex (mbRead p >>= jPred, mbRead t >>= jPred, mbRead n >>= jPred)
    parseVert (c:cs) = case c of
                        ' ' -> parsePos  $ (split ' ' . dropWhile (==' ')) cs
                        't' -> parseTex  $ (split ' ' . dropWhile (==' ')) cs
                        'n' -> parseNorm $ (split ' ' . dropWhile (==' ')) cs
    parsePos  [x,y,z] = addPos (Point (read x) (read y) (read z))
    parseTex  [u,v]   = addTex  (Vec  (read u) (read v)    0.0  )
    parseNorm [x,y,z] = addNorm (Vec  (read x) (read y) (read z))
    
    addPos p (ps,ts,ns,is) = (p:ps,ts,ns,is)
    addTex t (ps,ts,ns,is) = (ps,t:ts,ns,is)
    addNorm n (ps,ts,ns,is) = (ps,ts,n:ns,is)                     
    addIndex i (ps,ts,ns,is) = (ps,ts,ns,i:is)

resolve :: ([HVec Float], [HVec Float], [HVec Float], [(Maybe Int,Maybe Int,Maybe Int)]) -> Mesh
resolve = r.vectorise
  where
    vectorise (ps,ts,ns,fs) = (V.fromList ps, V.fromList ts,
                               V.fromList ns, reverse fs)
    r (ps,ts,ns,is) = Mesh ps ns ts $ map t is 

    t (p,t,n) = 
      case p of
       Just p' -> case t of
         Just t' -> case n of
           Just n' -> (p',n',t')
           Nothing -> error "unsupported obj file"
         Nothing -> error "unsupported obj file"
       Nothing -> error "unsupported obj file"
    
